import pytest
from finesse.symbols import OPERATORS, FUNCTIONS
from finesse.script.generator import ElementContainer
from ...util import dedent_multiline
from ..data import NONE, BOOLEANS, STRINGS, INTEGERS, FLOATS, IMAGINARIES, ARRAYS


@pytest.fixture
def unbuilder(
    unbuilder,
    fake_element_cls,
    fake_element_noparam_cls,
    fake_detector_cls,
    fake_analysis_cls,
):
    unbuilder.spec.register_element(
        ("fake_element",), {"setter": fake_element_cls, "getter": fake_element_cls}
    )
    unbuilder.spec.register_element(
        ("fake_noparam",),
        {"setter": fake_element_noparam_cls, "getter": fake_element_noparam_cls},
    )
    unbuilder.spec.register_element(
        ("fake_detector",), {"setter": fake_detector_cls, "getter": fake_detector_cls}
    )
    unbuilder.spec.register_analysis(
        ("fake_analysis",), {"setter": fake_analysis_cls, "getter": fake_analysis_cls}
    )
    unbuilder.spec.binary_operators["+"] = OPERATORS["__add__"]
    unbuilder.spec.binary_operators["-"] = OPERATORS["__sub__"]
    unbuilder.spec.unary_operators["+"] = FUNCTIONS["pos"]
    unbuilder.spec.unary_operators["-"] = FUNCTIONS["neg"]
    unbuilder.spec.expression_functions["sin"] = FUNCTIONS["sin"]
    unbuilder.spec.expression_functions["arctan2"] = FUNCTIONS["arctan2"]

    return unbuilder


@pytest.mark.parametrize(
    "_,expected,arg", NONE + BOOLEANS + STRINGS + INTEGERS + FLOATS + IMAGINARIES
)
def test_terminal_value(unbuilder, fake_element_noparam_cls, model, _, expected, arg):
    """Values like ints, floats, strings; those without additional dependencies."""
    script = unbuilder.unbuild(
        ElementContainer(fake_element_noparam_cls("myelement", arg))
    )
    assert script == f"fake_noparam myelement a={expected} b=none"


def test_reference(unbuilder, fake_element_cls, model):
    model.add(fake_element_cls("myel1"))
    model.add(fake_element_cls("myel2"))
    model.myel1.a = 1 - model.myel1.b.ref
    model.myel2.a = model.myel1.a.ref
    model.myel2.b = model.myel1.b.ref
    assert unbuilder.unbuild(model) == dedent_multiline(
        """
        fake_element myel1 a=(1-&myel1.b) b=none
        fake_element myel2 a=&myel1.a b=&myel1.b
        """
    )


@pytest.mark.parametrize(
    "function,expected",
    (
        (FUNCTIONS["neg"](1), "-1"),
        (FUNCTIONS["sin"](1), "sin(1)"),
        (FUNCTIONS["arctan2"](1, 2), "arctan2(1, 2)"),
    ),
)
def test_expression_function(unbuilder, fake_element_cls, model, function, expected):
    script = unbuilder.unbuild(ElementContainer(fake_element_cls("myel1", a=function)))
    assert script == f"fake_element myel1 a={expected} b=none"


@pytest.mark.parametrize("expected,array", ARRAYS)
def test_array(unbuilder, fake_element_noparam_cls, model, array, expected):
    script = unbuilder.unbuild(
        ElementContainer(fake_element_noparam_cls("myel1", a=array))
    )
    assert script == f"fake_noparam myel1 a={expected} b=none"


def test_port(unbuilder, fake_element_cls, fake_detector_cls, model):
    model.add(fake_element_cls("myel1"))
    model.add(fake_detector_cls("mydet1", target=model.myel1.p1))
    assert unbuilder.unbuild(model) == dedent_multiline(
        """
        fake_element myel1 a=none b=none
        fake_detector mydet1 target=myel1.p1
        """
    )


def test_node(unbuilder, fake_element_cls, fake_detector_cls, model):
    model.add(fake_element_cls("myel1"))
    model.add(fake_detector_cls("mydet1", target=model.myel1.p1.o))
    assert unbuilder.unbuild(model) == dedent_multiline(
        """
        fake_element myel1 a=none b=none
        fake_detector mydet1 target=myel1.p1.o
        """
    )


@pytest.mark.parametrize(
    "name,args,kwargs,expected",
    (
        ("myelement", [], {}, "fake_noparam myelement a=none b=none"),
        ("myelement", [None], {}, "fake_noparam myelement a=none b=none"),
        ("myelement", [None, None], {}, "fake_noparam myelement a=none b=none"),
        ("myelement", [None], {"b": None}, "fake_noparam myelement a=none b=none"),
        ("myelement", [], {"a": None}, "fake_noparam myelement a=none b=none"),
        ("myelement", [], {"b": None}, "fake_noparam myelement a=none b=none"),
        (
            "myelement",
            [],
            {"a": None, "b": None},
            "fake_noparam myelement a=none b=none",
        ),
        ("myelement", [1], {}, "fake_noparam myelement a=1 b=none"),
        ("myelement", [1, 2], {}, "fake_noparam myelement a=1 b=2"),
        ("myelement", [], {"a": 1}, "fake_noparam myelement a=1 b=none"),
        ("myelement", [], {"a": 1, "b": 2}, "fake_noparam myelement a=1 b=2"),
    ),
)
def test_element(
    unbuilder, fake_element_noparam_cls, model, name, args, kwargs, expected
):
    script = unbuilder.unbuild(
        ElementContainer(fake_element_noparam_cls(name, *args, **kwargs))
    )
    assert script == expected


@pytest.mark.parametrize(
    "args,kwargs,expected",
    (
        ([], {}, "fake_analysis(a=none, b=none)"),
        ([None], {}, "fake_analysis(a=none, b=none)"),
        ([None, None], {}, "fake_analysis(a=none, b=none)"),
        ([None], {"b": None}, "fake_analysis(a=none, b=none)"),
        ([], {"a": None}, "fake_analysis(a=none, b=none)"),
        ([], {"b": None}, "fake_analysis(a=none, b=none)"),
        ([], {"a": None, "b": None}, "fake_analysis(a=none, b=none)"),
        ([1], {}, "fake_analysis(a=1, b=none)"),
        ([1, 2], {}, "fake_analysis(a=1, b=2)"),
        ([], {"a": 1}, "fake_analysis(a=1, b=none)"),
        ([], {"a": 1, "b": 2}, "fake_analysis(a=1, b=2)"),
    ),
)
def test_analysis(unbuilder, fake_analysis_cls, model, args, kwargs, expected):
    script = unbuilder.unbuild(fake_analysis_cls(*args, **kwargs))
    assert script == expected



"""Kat file parser expression parsing fuzzing tests."""

from hypothesis import given, settings, note, assume
from ......util import recursive_expressions


# Some expressions produce errors that are hard to detect before evaluating, so we
# instead define some allowed failures.
ALLOWED_ERRORS = ["can't take floor of complex number.", "complex exponentiation"]


@given(value=recursive_expressions())
@settings(deadline=1000)  # See #292.
def test_expression_fuzzing(fuzz_value_parse_compare, value):
    note(f"Expression: {value}")

    try:
        # The str() method of Operation should be Python syntax compatible.
        expected = eval(str(value))
        fuzz_value_parse_compare(value, expected, exact=False)
    except ZeroDivisionError:
        # There are infinitely many zero division errors so we catch them before the
        # general catch-all exception below.
        assume(False)
    except Exception as e:
        if str(e) in ALLOWED_ERRORS:
            assume(False)
        else:
            raise



import pytest
import finesse
import numpy as np


def FT_GW_sidebands(Lambda, h0, fsig, L, n, sb_sign, P0):
    """Reference for strain signal in a space.

    Interferometer responses to gravitational waves:
    Comparing Finesse simulations and analytical solutions
    Charlotte Bond, Daniel Brown and Andreas Freise
    LIGO DCC: T1300190
    https://arxiv.org/pdf/1306.6752.pdf
    """
    # Carrier light parameters
    c = 299792458
    f0 = c / Lambda
    w0 = 2 * np.pi * f0
    # Signal anglar frequency
    wsig = 2 * np.pi * fsig
    # Sideband amplitude
    Asb = -np.sqrt(P0) * (w0 * h0 / (2 * wsig)) * np.sin(wsig * L * n / (2 * c))
    # Phase
    phi_sb = np.pi / 2 - sb_sign * wsig * L * n / (2 * c)
    # Final sideband
    return Asb * np.exp(1j * phi_sb)


@pytest.mark.parametrize(
    "L,P", ((1, 1), (2, 10), (100, 4), (3, 1000)),
)
def test_space_strain_signal(L, P):
    IFO = finesse.Model()

    IFO.parse(
        f"""
        l l1 P={P}
        s s1 l1.p1 n1.p1 L={L}
        nothing n1
        ad upper n1.p1.i &fsig.f
        ad lower n1.p1.i -&fsig.f
        sgen sig s1.h
        fsig(10)
        xaxis(fsig.f, lin, 1, 100k, 1000)
        """
    )

    out = IFO.run()

    upper_err = abs(
        out["upper"]
        - FT_GW_sidebands(
            IFO.lambda0,
            1,
            out.x1,
            IFO.spaces.s1.L.value,
            IFO.spaces.s1.nr.value,
            +1,
            IFO.l1.P.value,
        )
    )
    lower_err = abs(
        out["lower"]
        - FT_GW_sidebands(
            IFO.lambda0,
            1,
            out.x1,
            IFO.spaces.s1.L.value,
            IFO.spaces.s1.nr.value,
            -1,
            IFO.l1.P.value,
        )
    )

    assert max(upper_err) < 1e-6
    assert max(lower_err) < 1e-6



"""Test cases for reflected and transmitted power from a single beam splitter."""

import math
import pytest
from finesse.analysis import noxaxis
from finesse.components import Laser, Space, Beamsplitter
from finesse.detectors import PowerDetector


@pytest.fixture
def laser_and_beamsplitter(model):
    """Model with a laser and beamsplitter separated by a space."""
    model.chain(
        Laser("L0"), Space("s0"), Beamsplitter("BS", R=0.5, T=0.5),
    )
    model.add(PowerDetector("refl", model.BS.p2.o))
    model.add(PowerDetector("trns", model.BS.p3.o))
    return model


def test_fully_reflective(laser_and_beamsplitter):
    """Test fully reflective beamsplitter power."""
    laser_and_beamsplitter.BS.set_RTL(R=1, T=0)

    assert float(laser_and_beamsplitter.BS.T) == 0

    out = noxaxis(laser_and_beamsplitter)

    assert math.isclose(out["refl"], float(laser_and_beamsplitter.L0.P))
    assert out["trns"] == 0


def test_fully_transmissive(laser_and_beamsplitter):
    """Test fully transmissive beamsplitter power."""
    laser_and_beamsplitter.BS.set_RTL(T=1, R=0)

    assert float(laser_and_beamsplitter.BS.R) == 0

    out = noxaxis(laser_and_beamsplitter)

    assert math.isclose(out["trns"], float(laser_and_beamsplitter.L0.P))
    assert out["refl"] == 0


def test_half_reflective(laser_and_beamsplitter):
    """Test half reflective beamsplitter power."""
    laser_and_beamsplitter.BS.set_RTL(T=0.5, R=0.5)

    assert math.isclose(float(laser_and_beamsplitter.BS.T), 0.5)

    out = noxaxis(laser_and_beamsplitter)

    assert math.isclose(out["refl"], 0.5 * float(laser_and_beamsplitter.L0.P))
    assert math.isclose(out["trns"], 0.5 * float(laser_and_beamsplitter.L0.P))



import click
from .util import (
    input_file_argument,
    output_file_argument,
    legacy_option,
    verbose_option,
    quiet_option,
    fancy_error_option,
    debug_option,
    log_exclude_option,
    parse_path,
    KatState,
)


@click.command()
@input_file_argument
@output_file_argument
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def convert(ctx, input_file, output_file, legacy):
    """Convert and normalize kat script to canonical form.

    This can be used to convert Finesse 2 or Finesse 3 scripts to canonical Finesse 3 form. Note
    that some Finesse 2 instructions are not supported. The resulting script is arranged in standard
    order and with standard spacing.

    If OUTPUT_FILE is not specified, it defaults to stdout.
    """
    # Get state object to set verbosity, etc.
    state = ctx.ensure_object(KatState)

    model = parse_path(state, input_file, legacy=legacy)
    model.unparse_file(output_file)


if __name__ == "__main__":
    convert()



"""Test modes command parsing."""

import numpy as np


def test_modes(model):
    """Test that the modes command parses correctly."""
    model.parse("modes([[0, 1], [1, 0]])")
    assert np.all(model.homs == [[0, 1], [1, 0]])



import pytest
import finesse


@pytest.fixture
def solution():
    base = finesse.Model()
    base.parse(
        """
        l L0 P=1

        s s0 L0.p1 m1.p1

        m m1 R=0 T=1

        ad a00 m1.p2.o 0 n=0 m=0
        ad a10 m1.p2.o 0 n=1 m=0
        ad a01 m1.p2.o 0 n=0 m=1
        pd P m1.p2.o

        gauss g1 L0.p1.o w0=1m z=0

        modes(maxtem=1)
        """
    )

    base.L0.tem(0, 0, 1, 0)
    base.L0.tem(1, 0, 2, 0)
    base.L0.tem(0, 1, 4, 0)

    return base.run()


def test_power_conserved(solution):
    sol = solution
    assert abs(sol["P"] - 1) <= 1e-15


def test_hom_relative_power(solution):
    sol = solution
    assert abs(2 * abs(sol["a00"]) ** 2 - abs(sol["a10"]) ** 2) <= 1e-15
    assert abs(4 * abs(sol["a00"]) ** 2 - abs(sol["a01"]) ** 2) <= 1e-15



"""
Plane wave quantum noise tests.
"""

import pytest
import numpy as np
from scipy import constants
from numpy.testing import assert_allclose


@pytest.fixture
def lossy_mirror_model(model):
    model.parse(
        """
        l l1 P=1
        s s1 l1.p1 m1.p1 L=0
        m m1 R=0.5 T=0.4

        qnoised qd_laser l1.p1.o
        qnoised qd_trans m1.p2.o
        qnoised qd_refl m1.p1.o
    """
    )

    return model


@pytest.fixture
def modulator_model(model):
    model.parse(
        """
        l l1 P=1
        s s1 l1.p1 mod1.p1 L=0
        mod mod1 f=10M midx=0.8 mod_type=am

        qnoised qd1 mod1.p2.o
        qnoised1 qd2 mod1.p2.o &mod1.f 0

        fsig(1)
    """
    )

    return model


@pytest.fixture
def squeezer_model(model):
    model.parse(
        """
        sq sqz 10
        s s1 sqz.p1 mod.p1 L=5
        mod mod f=10M midx=0.5
        s s2 mod.p2 m1.p1 L=5
        m m1 R=0.9 T=0.1
        s scav m1.p2 m2.p1 L=30
        m m2 R=0.9 T=0.1

        free_mass fm m2.mech mass=1

        qnoised qd1 m1.p1.o
        qnoised1 qd2 m1.p1.o &mod.f 0
        qnoised1 qd3 m1.p1.o -&mod.f 0
        qnoised1 qd4 m1.p1.o &mod.f 0
        qnoised1 qd5 m1.p1.o -&mod.f 0

        fsig(1)
    """
    )

    return model


def test_vacuum_quantum_noise(lossy_mirror_model):
    """Test basic quantum noise generation"""
    lossy_mirror_model.parse(
        """
        xaxis(l1.f, log, 1, 1e9, 100)  # Sweep laser frequency from 1Hz to 1GHz
        """
    )

    # Trying to detect quantum noise without an fsig should raise an error
    with pytest.raises(Exception):
        lossy_mirror_model.run()
    lossy_mirror_model.parse("fsig(1)")
    out = lossy_mirror_model.run()
    xaxis = out.x[0]

    # Quantum noise sidebands generated by the laser
    qn = 0.5 * (1 + xaxis / lossy_mirror_model.f0)

    # Convert generated Qn to an ASD
    # Factors:
    #   2 (number of sidebands being added)
    # * laser_power
    # * 4 (Where does this come from?)
    # * 2 (compensation for 0.5 demod factor when demodulating at signal frequency)
    # * 0.25 (factor 0.25 per demodulation)
    # TODO: Won't the 4 and 0.25 always cancel each other out? What's
    # the justification for these?
    fac = 2 * lossy_mirror_model.l1.P.value * 4 * 2 * 0.25
    qn_asd = np.sqrt(fac * qn * constants.h * lossy_mirror_model.f0)
    qn_asd_trans = qn_asd * np.sqrt(lossy_mirror_model.m1.T.value)
    qn_asd_refl = qn_asd * np.sqrt(lossy_mirror_model.m1.R.value)

    assert_allclose(out["qd_laser"], qn_asd, rtol=1e-14, atol=0)
    assert_allclose(out["qd_trans"], qn_asd_trans, rtol=1e-14, atol=0)
    assert_allclose(out["qd_refl"], qn_asd_refl, rtol=1e-14, atol=0)


def test_modulator_quantum_noise_sweep_laser_freq(modulator_model):
    """
    Test modulator quantum noise couplings with varying laser frequency
    """
    modulator_model.parse(
        """
        xaxis(l1.f, log, 1, 1e9, 100)  # Sweep laser frequency from 1Hz to 1GHz
        """
    )

    out = modulator_model.run()
    xaxis = out.x[0]

    # Quantum noise sidebands generated by the laser
    qn = 0.5 * (1 + xaxis / modulator_model.f0)

    # Combine the quantum noise from each carrier with the correct modulated powers
    midx = modulator_model.mod1.midx.value
    qn_tot = qn * ((1 - 0.5 * midx) ** 2 + 2 * (0.25 * midx) ** 2)

    # Convert generated qn to an ASD
    # Factors:
    #   2 (number of sidebands being added)
    # * laser_power
    # * 4 (Where does this come from?)
    # * 2 (compensation for 0.5 demod factor when demodulating at signal frequency)
    # * 0.25 (factor 0.25 per demodulation)
    # TODO: Won't the 4 and 0.25 always cancel each other out? What's
    # the justification for these?
    fac = 2 * modulator_model.l1.P.value * 4 * 2 * 0.25
    qn_asd = np.sqrt(fac * qn_tot * constants.h * modulator_model.f0)

    assert_allclose(out["qd1"], qn_asd, rtol=1e-14, atol=0)

    # TODO: work out the correct analytics for this
    # assert_allclose(out["qd2"], qn_asd_upper, rtol=1e-14, atol=0)


def test_modulator_quantum_noise_sweep_modulation_index(modulator_model):
    """
    Test modulator quantum noise couplings with varying modulation index
    """
    modulator_model.parse(
        """
        xaxis(mod1.midx, lin, 0, 1, 100)
        """
    )

    out = modulator_model.run()
    midx = out.x[0]

    # Quantum noise sidebands generated by the laser
    qn = 0.5

    # Combine the quantum noise from each carrier with the correct modulated powers
    qn_tot = qn * ((1 - 0.5 * midx) ** 2 + 2 * (0.25 * midx) ** 2)

    # Convert generated qn to an ASD
    # Factors:
    #   2 (number of sidebands being added)
    # * laser_power
    # * 4 (Where does this come from?)
    # * 2 (compensation for 0.5 demod factor when demodulating at signal frequency)
    # * 0.25 (factor 0.25 per demodulation)
    # TODO: Won't the 4 and 0.25 always cancel each other out? What's
    # the justification for these?
    fac = 2 * modulator_model.l1.P.value * 4 * 2 * 0.25
    qn_asd = np.sqrt(fac * qn_tot * constants.h * modulator_model.f0)

    assert_allclose(out["qd1"], qn_asd, rtol=1e-14)

    # TODO: work out the correct analytics for this
    # assert_allclose(out["qd2"], qn_asd_upper, rtol=1e-14, atol=0)


@pytest.mark.skip(reason="Not valid until finesse 3 handles this properly")
def test_modulator_quantum_noise_squeezer(squeezer_model):
    """Test modulator quantum noise couplings with squeezed light input"""
    squeezer_model.parse(
        """
        xaxis(sqz.f, lin, -10, 10, 20)  # Sweep squeezer frequency from -10Hz to +10Hz
        """
    )

    out = squeezer_model.run()
    xaxis = out.x[0]
    qn = 0.5 * (1 + xaxis / squeezer_model.f0) * constants.h

    assert_allclose(out["qd1"], qn, rtol=1e-5, atol=1e-8)
    assert_allclose(out["qd2"], qn, rtol=1e-5, atol=1e-8)
    assert_allclose(out["qd3"], qn, rtol=1e-5, atol=1e-8)



import pytest
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler, fake_binop_cls):
    element = lambda a=None, b=None, c=None, d=None: None
    compiler.spec.register_element(
        "fake_element", {"setter": element, "getter": element}
    )
    compiler.spec.binary_operators["+"] = lambda a, b: fake_binop_cls("+", a, b)
    return compiler


@pytest.mark.parametrize(
    "script,graph_definition",
    (
        pytest.param(
            "fake_element el1 1+2",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_element'"
                        name_token "'el1'"
                        type "ELEMENT"
                        extra_tokens "[' ', ' ']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        token "'+'"
                        type "EXPRESSION"
                    ]
                    node [
                        id 3
                        label ".script.0.0.0"
                        token "'1'"
                        type "VALUE"
                    ]
                    node [
                        id 4
                        label ".script.0.0.1"
                        token "'2'"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 2
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 4
                        target 2
                        type "ARGUMENT"
                        order "1"
                    ]
                ]
                """
            ),
            id="1+2",
        ),
        pytest.param(
            "fake_element el1 (1)",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_element'"
                        name_token "'el1'"
                        type "ELEMENT"
                        extra_tokens "[' ', ' ']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        type "GROUPED_EXPRESSION"
                        extra_tokens "['(', ')']"
                    ]
                    node [
                        id 3
                        label ".script.0.0.0"
                        token "'1'"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 2
                        type "ARGUMENT"
                        order "0"
                    ]
                ]
                """
            ),
            id="grouped expression",
        ),
    ),
)
def test_expression(assert_graphs_match, script, graph_definition):
    assert_graphs_match(script, graph_definition)



import inspect
import fnmatch
import click
from .util import KatState


@click.command()
@click.argument("query", required=False, nargs=-1)
@click.option(
    "--elements",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show only elements.",
)
@click.option(
    "--commands",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show only commands.",
)
@click.option(
    "--analyses",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show only analyses.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show extended documentation for each directive.",
)
@click.pass_context
def syntax(ctx, query, elements, commands, analyses, verbose):
    """Query the kat script syntax documentation.

    \b
    Simple wildcards are supported:

    - ``*`` matches 0 or more characters
    - ``?`` matches any single character
    - ``[abc]`` matches any characters in abc
    - ``[!abc]`` matches any characters not in abc

    Specify zero or more QUERY terms. By default, all syntax is shown.
    """
    from ..script import syntax
    from ..script.spec import KatSpec
    from ..script.adapter import ElementAdapter, CommandAdapter, AnalysisAdapter

    state = ctx.ensure_object(KatState)
    spec = KatSpec()

    if not elements and not commands and not analyses:
        directives = spec.directives
    else:
        directives = {}

        if elements:
            directives.update(spec.elements)
        if commands:
            directives.update(spec.commands)
        if analyses:
            directives.update(spec.analyses)

    adapters = set()
    for directive, adapter in directives.items():
        if query and not any(fnmatch.fnmatch(directive, q) for q in query):
            continue

        adapters.add(adapter)

    if not adapters:
        state.print_error("No directives found.")

    # Default colours for each directive type.
    colors = {ElementAdapter: "green", CommandAdapter: "yellow", AnalysisAdapter: "red"}

    # Sort the directives in order they appear in the spec. This is better than alphabetic sort,
    # which puts `x2axis` ahead of `xaxis`.
    order = list(directives)
    for adapter in sorted(
        adapters, key=lambda adapter: order.index(adapter.short_name)
    ):
        syntax_form = click.style(syntax(adapter.short_name))
        directive = " / ".join(sorted(adapter.aliases, key=lambda alias: len(alias)))
        fg = colors[type(adapter)]

        if verbose:
            if docstring := inspect.getdoc(adapter.docobj_type):
                # Grab only the first line of the docstring.
                doc = docstring.splitlines()[0]
            else:
                doc = "(no documentation)"

            state.print(directive, fg=fg)
            state.print("=" * len(directive), fg=fg)
            state.print(doc)
            state.print()
            state.print(syntax_form)
            state.print()
        else:
            state.print(f"{click.style(directive, fg=fg)}: {syntax_form}")


if __name__ == "__main__":
    syntax()



import pytest
from finesse.script.compiler import KatCompiler
from ....diff import assert_graph_matches_def


@pytest.fixture
def compiler(spec):
    return KatCompiler(spec=spec)


@pytest.fixture
def assert_graphs_match(compiler):
    def matcher(script, reference_graph):
        compiler.compile(script)
        assert_graph_matches_def(compiler.graph, reference_graph)

    return matcher



"""Mirror unit tests."""

import pytest
import numpy as np
from finesse.components import Mirror
from . import TestSurface


@pytest.fixture
def component():
    """Set the fixture used by TestMirrorSurface to a mirror."""
    return Mirror


class TestMirrorSurface(TestSurface):
    """Test surface properties (R, T, L, Rc, etc.)."""


@pytest.mark.parametrize(
    "Rcx,Rcy",
    (
        (np.inf, np.inf),
        (np.inf, -np.inf),
        (-np.inf, np.inf),
        (-np.inf, -np.inf),
        (1, 1),
        (1, -1),
        (-1, 1),
        (-1, -1),
        (3.141, 6.282),
        (3.141, -6.282),
        (-3.141, 6.282),
        (-3.141, -6.282),
        (1e9, 1e-6),
        (1e9, -1e-6),
        (-1e9, 1e-6),
        (-1e9, -1e-6),
    ),
)
def test_abcd(Rcx, Rcy):
    """Test that the ABCD matrix of particular combinations of ports shows symmetry with other \
    ports.

    The sign of Rc matters when we test against port pairs on the opposite surface. See
    https://finesse.readthedocs.io/en/latest/api/components/mirror/generated/finesse.components.mirror.Mirror.ABCD.html#finesse.components.mirror.Mirror.ABCD.

    For REFLECTION:

        Mirror( Rcx,  Rcy).p1.i -> Mirror( Rcx,  Rcy).p1.o

                            is equal to:

        Mirror(-Rcx, -Rcy).p2.i -> Mirror(-Rcx, -Rcy).p2.o

    For TRANSMISSION:

        Mirror( Rcx,  Rcy).p1.i -> Mirror( Rcx,  Rcy).p2.o

                            is equal to:

        Mirror(-Rcx, -Rcy).p2.i -> Mirror(-Rcx, -Rcy).p1.o
    """

    m_ref = Mirror("m", Rc=(Rcx, Rcy))
    m_cmp = Mirror("m", Rc=(-Rcx, -Rcy))

    ## Each test below checks x- then y-directions.

    # Reflection
    r_lhs_x = m_ref.ABCD(m_ref.p1.i, m_ref.p1.o, direction="x")
    r_lhs_y = m_ref.ABCD(m_ref.p1.i, m_ref.p1.o, direction="y")

    assert np.all(r_lhs_x == m_cmp.ABCD(m_cmp.p2.i, m_cmp.p2.o, direction="x"))
    assert np.all(r_lhs_y == m_cmp.ABCD(m_cmp.p2.i, m_cmp.p2.o, direction="y"))

    # Transmission
    t_lhs_x = m_ref.ABCD(m_ref.p1.i, m_ref.p2.o, direction="x")
    t_lhs_y = m_ref.ABCD(m_ref.p1.i, m_ref.p2.o, direction="y")

    # (1)
    assert np.all(t_lhs_x == m_cmp.ABCD(m_cmp.p2.i, m_cmp.p1.o, direction="x"))
    assert np.all(t_lhs_y == m_cmp.ABCD(m_cmp.p2.i, m_cmp.p1.o, direction="y"))



"""Symbol fuzzing tests."""

import math
import cmath
from hypothesis import given
from hypothesis.strategies import floats, complex_numbers
from finesse.symbols import FUNCTIONS


@given(number=floats(allow_infinity=False, allow_nan=False))
def test_expression_fuzzing_with_floats(number):
    """Test that expressions with numbers evaluate to the correct values."""
    assert math.isclose(FUNCTIONS["cos"](number).eval(), math.cos(number))


# Maximum magnitude is set to avoid math range errors and infinities.
@given(number=complex_numbers(max_magnitude=1e2))
def test_expression_fuzzing_with_complex_numbers(number):
    """Test that expressions with numbers evaluate to the correct values."""
    assert cmath.isclose(FUNCTIONS["exp"](number).eval(), cmath.exp(number))



"""Tools for calculating coupling coefficients and modal scattering matrices of
different types."""

from .bayerhelms import make_bayerhelms_matrix


# TODO (sjr) Create functions corresponding to different scattering
#            matrix types as they are implemented in the code (i.e.
#            things like aperture and map knm matrix computations)
def make_scatter_matrix(mtype, *args, **kwargs):
    """Constructs and computes a coupling coefficient scattering matrix of the specified
    type (via `mtype`).

    Parameters
    ----------
    mtype : str
        Type of scattering matrix to compute.

        The valid options are:

          - "bayerhelms" --- computes a Bayer-Helms scattering matrix,
                             see :func:`.make_bayerhelms_matrix`.

    *args : positional arguments
        The positional arguments to pass to the relevant scatter matrix function.

    **kwargs : keyword arguments
        The keyword arguments to pass to the relevant scatter matrix function.

    Returns
    -------
    kmat : :class:`.KnmMatrix`
        The resulting scattering matrix as a :class:`.KnmMatrix` object.

    Examples
    --------
    See :ref:`arbitrary_scatter_matrices`.
    """
    mtype_to_func = {"bayerhelms": make_bayerhelms_matrix}

    func = mtype_to_func.get(mtype.casefold())
    if func is None:
        raise ValueError(f"Unrecognised / not-yet-implemented mtype argument: {mtype}")

    return func(*args, **kwargs)
